<?php

/*
Формат вызова API:

1. Получить значение свойства:
	http://webserver/api.php?a=DEVICETYPE:DEVICE:PROPERTY
2. Установить значение свойства:
	http://webserver/api.php?a=DEVICETYPE:DEVICE:PROPERTY=VALUE
3. Вызвать метод:
	http://webserver/api.php?a=DEVICETYPE:DEVICE:METHOD
   Параметры вызова передаются закодированные JSON-ом в параметре $_REQUEST['p']

4. проверить наличие обновлений
	http://webserver/api.php?a=checkupdate
5. Обновить до последней версии
	http://webserver/api.php?a=update

где
  DEVICETYPE - тип устройства, к которому обращаемся (fiscalprinter, cashdrawer, display)
  DEVICE - наименование конкретного устройства. Например, для ФР ШТРИХ-ФР-К указываем shtrihm
  PROPERTY_OR_METHOD - имя свойства или метода в классе-обработчике устройства


К любому запросу можно приписать параметр:
	encode=TEXT для вывода ответа в текстовом виде (по умолчанию установлен JSON)
	loglevel=1,2,3 - уровень логгирования
	com=COM_NUM:BAUDRATE:MODEL

*/

DEFINE('API_VERSION_FILENAME', "version.inc");
global $API_VERSION;
$API_VERSION = 0;
if (file_exists(API_VERSION_FILENAME)) {
    $API_VERSION = file_get_contents(API_VERSION_FILENAME);
}

DEFINE('API_UPDATE_URL', 'https://core.virtualpos.ru/updates');

header('Content-Type: text/html; charset=utf-8');
header("Access-Control-Allow-Origin: *");

// откллючение кеширования info: http://stackoverflow.com/questions/49547/making-sure-a-web-page-is-not-cached-across-all-browsers
header('Cache-Control: no-cache, no-store, must-revalidate'); // HTTP 1.1.
header('Pragma: no-cache'); // HTTP 1.0.
header('Expires: 0'); // Proxies.

DEFINE('ROOT_DIR', __DIR__);

error_reporting(E_ALL);
ini_set('display_errors', false);
set_time_limit(600);
register_shutdown_function("fatal_handler");

require 'core/config.php';
require_once ROOT_DIR . "/core/Log.php";//подключаем класс работы с логом

function fatal_handler()
{
    $error = error_get_last();
    if ($error !== NULL) {

        $errtype = $error["type"];
        $errfile = $error["file"];
        $errline = $error["line"];
        $errstr = $error["message"];

        $log = new Log('./logs/log.json');
        $log->set($errtype, $errfile, $errline, $errstr);

        if ($errtype === E_ERROR) {
            HWBridge::doExit(false, "Fatal error: $errstr in $errfile on line $errline");
        } else {
            $errLvl = error_reporting();
            if ($errtype <= $errLvl) {
                //echo ("Error (type $errtype) $errstr in $errfile on line $errline<br>");
            }
            error_log("Error (type $errtype) $errstr in $errfile on line $errline");
        }
    }
}

class HWBridge
{

    public static $encodeMethod = 'JSON'; //JSON, TEXT

    /**
     * Список зарегистрированных типов устройств, с которыми модуль умеет работать
     *
     */
    protected $REGISTERED_DEVICES = array('fiscalprinter', 'cashdrawer', 'display', 'pinpad', 'egais', 'scale');

    const ENABLE_AUTO_UPDATE_DEFAULT = 1;
    const ENABLE_AUTO_UPDATE = 'enable_auto_update';

    /**
     *  Функция выдает ответ в стандартном для АПИ виде и завершает работу приложения
     *
     * @param bool $isSuccess успешное завершение или нет. Если $message это архив, и в нем уже есть элемент success, то его значение не меняется
     * @param mixed $message доп. информация
     *
     * return void
     */
    public static function doExit($isSuccess, $message)
    {
        if (is_scalar($message)) {
            $result = array('success' => $isSuccess, 'text' => $message);
        } elseif (is_array($message)) {
            $result = $message;
            if (!isset($result['success'])) $result['success'] = $isSuccess;
        } else {
            $result = array('success' => $isSuccess);
        }
        if (self::$encodeMethod == 'JSON') {
            $result = json_encode($result, JSON_UNESCAPED_UNICODE);
        } else {
            $result = print_r($result, true);
        }
        echo $result;
        exit();
    }

    /**
     *  Принимает на входе строку вида: DEVICE_TYPE:DEVICE:METHOD_OR_PROPERTY
     *  и возвращает массив с отдельными элементами
     *
     */
    protected function parseDeviceModule($devicePath)
    {
        $devicePathItems = explode(':', $devicePath);
        if (count($devicePathItems) != 3) $this->doExit(false, array('error' => "Неверное обращение к устройству [$devicePath]"));

        $device = $devicePathItems[0];
        $model = $devicePathItems[1];
        $command = $devicePathItems[2];

        if (!in_array($device, $this->REGISTERED_DEVICES)) $this->doExit(false, array('error' => "Тип устройства [$device] не зарегистрирован в системе"));
        if (!file_exists($device)) $this->doExit(false, array('error' => "Тип устройства [$device] не найден"));
        if (!file_exists("$device/$model/{$device}_$model.php")) $this->doExit(false, array('error' => "Модель устройства [$model] не зарегистрирована в системе"));

        return array(
            'modulename' => "$device/$model/{$device}_$model.php",
            'classname' => $device . '_' . $model,
            'parent_classname' => $device,
            'command' => $command
        );
    }

    /**
     * Основная процедура обработки API вызовов
     *
     */
    public function Processing()
    {
        $isMethodCall = NULL;
        $params = array();
        global $API_VERSION;

        if (isset($_REQUEST['encode']) && ($_REQUEST['encode'] == 'TEXT')) self::$encodeMethod = 'TEXT';
        if (!isset($_GET['a'])) $this->doExit(false, array('error' => 'Не передан обязательный параметр'));
        $loglevel = (isset($_REQUEST['loglevel']) && in_array($_REQUEST['loglevel'], array('0', '1', '2', '3'))) ? $_REQUEST['loglevel'] : 0;

        //Запрос на автообновление
        if ($_GET['a'] == 'checkupdate') {
            return $this->update(true, $loglevel != 0);
        } elseif ($_GET['a'] == 'update') {
            return $this->update(false, $loglevel != 0);
        } elseif ($_GET['a'] == 'hello') {
            $this->doExit(true, array('message' => "VirtualPOS Hardware Bridge", 'version' => $API_VERSION));
        } elseif ($_GET['a'] == 'log') {
            $log = new Log(ROOT_DIR . './logs/log.json');
            $this->doExit(true, array('data' => $log->getAll(true), 'version' => $API_VERSION));
        }

        $comPortNumber = -1;
        $comPortBaud = -1;
        $model = -1;
        $ipAddress = -1;
        $tcpPort = -1;
        //В явном виде был передан com-порт
        if (isset($_REQUEST['com'])) {
            $comPortItems = explode(':', $_REQUEST['com'], 3);
            if (count($comPortItems) == 1) {
                $this->doExit(false, array('error' => "Адрес ФР задан неверно"));
            }
            $ipOctets = explode('.', $comPortItems[0]);
            if (count($ipOctets) == 4) {
                $ipAddress = $comPortItems[0];
                $tcpPort = $comPortItems[1];
            } else {
                $comPortNumber = $comPortItems[0];
                if (!is_numeric($comPortNumber))
                    $this->doExit(false, array('error' => "Неверное значение для COM порта {$comPortItems[0]}"));
                if (isset($comPortItems[1])) {
                    $comPortBaud = $comPortItems[1];
                    if (!is_numeric($comPortBaud))
                        $this->doExit(false, array('error' => "Неверное значение для скорости порта COM{$comPortItems[0]} - [{$comPortBaud}]"));
                }
                if (isset($comPortItems[2])) {
                    $model = $comPortItems[2];
                }
            }
        }

        // если пришел тест чека
        if (isset($_REQUEST['test_check'])) {
            $_REQUEST['p'] = $this->getTestCheckParam();
        }

        //Обращение к свойству или методу объекта
        $input = explode('=', $_GET['a'], 2);
        $devicepath = $input[0];
        if (count($input) == 2) {
            //Формат PROPERTY=VALUE
            $params = array('value' => $input[1]);
            $isMethodCall = false;
        } else {
            //В данном случае пока неясно, это запрос значения свойства или это вызов метода
            //Если переданы параметры, то это однозначно вызов метода
            if (isset($_REQUEST['p'])) {
                $isMethodCall = true;
                $params = json_decode($_REQUEST['p'], true);
                if (!isset($params)) $this->doExit(false, array('error' => "Не удалось декодировать параметры вызова"));
            }
        }

        //Парсим переданный вызов
        $module = $this->parseDeviceModule($devicepath);
        $CLASS_NAME = $module['classname'];
        $PARENT_CLASS_NAME = $module['parent_classname'];
        $command = $module['command'];

        //Подключаем модуль-обработчик
        require $module['modulename'];
        if (!class_exists($CLASS_NAME)) $this->doExit(false, array('error' => "Не найден класс-обработчик [$CLASS_NAME]"));
        if (!is_subclass_of($CLASS_NAME, $PARENT_CLASS_NAME)) $this->doExit(false, array('error' => "Класс-обработчик [$CLASS_NAME] должен быть унаследован от базового класса [$PARENT_CLASS_NAME]"));

        //Выясняем окончательно, запрашивается ли у нас свойство или происходит вызов метода
        if (!isset($isMethodCall)) $isMethodCall = !property_exists($CLASS_NAME, $command);


        if ($isMethodCall) {
            /*Вызываем запрошенный метод*/
            if (!method_exists($CLASS_NAME, $command)) $this->doExit(false, array('error' => "В классе [$CLASS_NAME] отсутствует метод [$command]"));
            $deviceObject = new $CLASS_NAME($loglevel);
            $deviceObject->SetComPort($comPortNumber);
            $deviceObject->SetComPortBaud($comPortBaud);
            $deviceObject->SetModel($model);
            $deviceObject->SetIpAddress($ipAddress);
            $deviceObject->SetTcpPort($tcpPort);
            try {
                $result = $deviceObject->$command($params);
            } catch (Exception $e) {
                $this->doExit(false, array('error' => $e->getMessage()));
            }
            $this->doExit(true, $result);
        } else {
            /**/
            if (!property_exists($CLASS_NAME, $command)) $this->doExit(false, array('error' => "В классе [$CLASS_NAME] отсутствует свойство [$command]"));
            $deviceObject = new $CLASS_NAME($loglevel);
            $deviceObject->SetComPort($comPortNumber);
            $deviceObject->SetComPortBaud($comPortBaud);
            $deviceObject->SetModel($model);

            if (isset($params['value'])) {
                /*Устанавливаем новое значение для свойства*/
                try {
                    $deviceObject->$command = $params['value'];
                } catch (Exception $e) {
                    $this->doExit(false, array('error' => $e->getMessage()));
                }
                $result = $deviceObject->$command;
                $this->doExit(true, array('message' => 'OK', $command => $result));
            } else {
                /*Получаем текущее значение свойства*/
                try {
                    $result = $deviceObject->$command;
                } catch (Exception $e) {
                    $this->doExit(false, array('error' => $e->getMessage()));
                }
                $this->doExit(true, array('message' => 'OK', $command => $result));
            }
        }

    }

    /**
     * Автоматическое обновление всех модулей.
     *
     * @param bool $justCheck если истина, то выполняется только проверка наличия новой версии. Само обновление не скачивается и не устанавливается
     * @return
     */
    public function update($justCheck = true, $log = false)
    {
        $config = Config::getInstance();
        $enableAutoUpdate = $config->getParameter(self::ENABLE_AUTO_UPDATE, self::ENABLE_AUTO_UPDATE_DEFAULT);
        if ($enableAutoUpdate == 0) {
            $this->doExit(true, array('needupdate' => false));
            return;
        }

        require('core/update.php');

        global $API_VERSION;
        $update = new AutoUpdate($log); //logging
        $update->currentVersion = $API_VERSION;
        $update->updateUrl = API_UPDATE_URL;

        $latest = $update->checkUpdate();
        if ($latest === false) $this->doExit(false, $update->getLastError());

        if ($justCheck || ($latest <= $update->currentVersion)) {
            $this->doExit(true, array(
                'needupdate' => ($latest > $update->currentVersion),
                'currentversion' => $update->currentVersion,
                'lastversion' => $latest, 'lastversionname' => $update->latestVersionName
            ));
        }

        if ($update->update()) {
            $this->doExit(true, array(
                'currentversion' => $update->currentVersion,
                'lastversion' => $latest, 'lastversionname' => $update->latestVersionName
            ));
        } else {
            $this->doExit(false, array(
                'text' => 'Ошибка. ' . $update->getLastError(),
                'currentversion' => $update->currentVersion,
                'lastversion' => $latest, 'lastversionname' => $update->latestVersionName
            ));
        }
    }

    private function getTestCheckParam()
    {
        return '{"meta":{"continue":false},"lines":[{"action":"print","params":{"string":"             ВиртуалПос             "}},{"action":"print","params":{"string":"         СПАСИБО ЗА ПОКУПКУ!        "}},{"action":"opencheck","params":{"password":30,"checktype":0}},{"action":"sale","params":{"item":"Рунебаунт (Runebound),","price":2030,"quantity":1}},{"action":"sale","params":{"item":"Memoir44: Breakthrough (Воспоминания о 1944: Прорыв)","price":1940,"quantity":3}},{"action":"sale","params":{"item":"Процветание (Prosperity)","price":2120,"quantity":3}},{"action":"sale","params":{"item":"Свинтус 3D","price":950,"quantity":1}},{"action":"printex","params":{"string":"------------------------------------"}},{"action":"printex","params":{"string":"Всего наименований: @TOTAL_OPERATIONS"}},{"action":"printex","params":{"string":"Касса: @POS_NUMBER Номер чека: @CURRENT_CHECK_NUMBER"}},{"action":"closecheck","params":{"payment":{"cash":"15160.00","card":0,"tare":0,"credit":0},"checkamount":"15160.00"}},{"action":"feed","params":{"linesQuantity":5}},{"action":"cut","params":{"timeout":3000}}]}';
    }
}


$bridge = new HWBridge();
$bridge->Processing();

?>
