<?php

/*
 * Set the script max execution time
 *
 * https://github.com/VisualAppeal/PHP-Auto-Update
 *
 */
ini_set('max_execution_time', 60);

define('UPDATE_DIR_TEMP', dirname(__FILE__).'/../temp/');
define('UPDATE_DIR_INSTALL', dirname(__FILE__).'/../');

require_once __DIR__ . "/Log.php";//подключаем класс работы с логом

class AutoUpdate {
	/*
	 * Enable logging
	 */
	private $_log = false;
	
	/* 
	 * Log file
	 */
	public $logDir  = 'logs';
	public $logFile = 'update.log';
	
	/*
	 * The last error
	 */
	private $_lastError = null;
	
	/*
	 * Current version
	 */
	public $currentVersion = 0;
	
	/*
	 * Name of the latest version
	 */
	public $latestVersionName = '';
	
	/*
	 * The latest version
	 */
	public $latestVersion = null;
	
	/*
	 * Url to the latest version of the update
	 */
	public $latestUpdate = null;
	
	/*
	 * Url to the update folder on the server
	 */
	public $updateUrl = 'http://example.com/updates/';
	
	/*
	 * Version filename on the server
	 */
	public $updateIni = 'update.ini';
	
	/*
	 * Temporary download directory
	 */
	public $tempDir = UPDATE_DIR_TEMP;
	
	/*
	 * Remove temprary directory after installation
	 */
	public $removeTempDir = true;
	
	/*
	 * Install directory
	 */
	public $installDir = UPDATE_DIR_INSTALL;
	
	/*
	 * Create new folders with this privileges
	 */
	public $dirPermissions = 0755;
	
	/*
	 * Update script filename
	 */
	public $updateScriptName = '_upgrade.php';
	
	/*
	 * Create new instance
	 *
	 * @param bool $log Default: false
	 */
	public function __construct($log = false) {
		$this->_log = $log;
	}
	
	/* 
	 * Log a message if logging is enabled
	 *
	 * @param string $message The message
	 *
	 * @return void
	 */
	public function log($message) {
		$errtype = 0;
		$errfile = 'update';
		$errline = '';
		$errstr  =  $message;

		//$log = new Log(__DIR__.'./../logs/log.json');
		//$log->set($errtype, $errfile, $errline, $errstr);

		$this->_lastError = $message;
		//echo ($this->_log ? 'log' : 'nolog');//$this->logFile . '<br/>';
		if ($this->_log) {
			if (!is_dir($this->logDir)) {
				if (!mkdir($this->logDir, 0777, true)) {
					error_log('Не удалось создать директорию `'.$this->logDir.'`!');
				}
			}

			$log = fopen($this->logDir.'/'.$this->logFile, 'a');
			if ($log) {
				$message = date('<Y-m-d H:i:s>').$message."\n";
				fputs($log, $message);
				fclose($log);
			}
			else {
				error_log('UPDATE MODULE. Could not write log file!. Message: ' . $message);
			}
		}
	}
	
	/*
	 * Get the latest error
	 *
	 * @return string Last error
	 */
	public function getLastError() {
		if (!is_null($this->_lastError))
			return $this->_lastError;
		else
			return false;
	}
	
	private function _removeDir($dir) {
		if (is_dir($dir)) {
			$objects = scandir($dir);
			foreach ($objects as $object) {
				if ($object != "." && $object != "..") {
					if (filetype($dir."/".$object) == "dir") 
						$this->_removeDir($dir."/".$object); 
					else 
						unlink($dir."/".$object);
				}
			}
			reset($objects);
			rmdir($dir);
		}
	}
	
	/*
	 * Check for a new version
	 *
	 * @return string The latest version
	 */
	public function checkUpdate() {
		$this->log('Проверка обновления. . .');
		
		$updateFile = $this->updateUrl.'/update.ini';
		
		$update = @file_get_contents($updateFile);
		if ($update === false) {
			$this->log('Не удалось получить файл обновления ['.$updateFile.']');
			return false;
		}
		else {
			$versions = parse_ini_string($update, true);
			if (is_array($versions)) {
				$keyOld = 0;
				$latest = 0;
				$update = '';
				
				foreach ($versions as $key => $version) {
					if ($key > $keyOld) {
						$keyOld = $key;
						$latest = $version['version'];
						$update = $version['url']; 
					}
				}
				
				$this->log('Обнаружена новая версия ['.$latest.'].');
				$this->latestVersion = $keyOld;
				$this->latestVersionName = $latest;
				$this->latestUpdate = $update;
				
				return $keyOld;
			}
			else {
				$this->log('Не удалось выполнить парсинг файла обновления');
				return false;
			}
		}
	}
	
	/*
	 * Download the update
	 *
	 * @param string $updateUrl Url where to download from
	 * @param string $updateFile Path where to save the download
	 */
	public function downloadUpdate($updateUrl, $updateFile) {
		$this->log('Скачивание обновления...');
		$update = @file_get_contents($updateUrl);
		
		if ($update === false) {
			$this->log('Не удалось скачать файл обновления '.$updateUrl);
			return false;
		}
		
		$handle = fopen($updateFile, 'w');
		
		if (!$handle) {
			$this->log('Не удалось сохранить скачанный файл обновления '.$updateFile);
			return false;
		}
		
		if (!fwrite($handle, $update)) {
			$this->log('Не удалось записать в файл обновления '.$updateFile);
			return false;
		}
		
		fclose($handle);
		
		return true;
	}
	
	/*
	 * Install update
	 *
	 * @param string $updateFile Path to the update file
	 */
	public function install($updateFile) {
		$zip = zip_open($updateFile);
			
		while ($file = zip_read($zip)) {				
			$filename = zip_entry_name($file);
			$foldername = $this->installDir.dirname($filename);
			
			$this->log('Обновление `'.$filename.'`!');
			
			if (!is_dir($foldername)) {
				if (!mkdir($foldername, $this->dirPermissions, true)) {
					$this->log('Не удалось создать директорию `'.$foldername.'`!');
				}
			}
			
			$contents = zip_entry_read($file, zip_entry_filesize($file));
			
			//Skip if entry is a directory
			if (substr($filename, -1, 1) == '/')
				continue;
			
			//Write to file
			if (file_exists($this->installDir.$filename) && !is_writable($this->installDir.$filename)) {
				$this->log('Не удалось обновить файл `'.$this->installDir.$filename.'` т.к. он недоступен для записи');
				return false;
			}
			
			$updateHandle = @fopen($this->installDir.$filename, 'w');
			
			if (!$updateHandle) {
				$this->log('Не удалось обновить файл `'.$this->installDir.$filename.'`');
				return false;
			}
			
			if (!fwrite($updateHandle, $contents)) {
				$this->log('Не удалось записать данные в файл `'.$this->installDir.$filename.'`');
				return false;
			}
			
			fclose($updateHandle);
			
			//If file is a update script, include
			if ($filename == $this->updateScriptName) {
				$this->log('Пытаемся подключить файл со сценарием обновления `'.$this->installDir.$filename.'`.');
				require($this->installDir.$filename);
				$this->log('Файл сценария обновления `'.$this->installDir.$filename.'` успешно подключен');
				unlink($this->installDir.$filename);
			}
		}
		
		zip_close($zip);
		
		if ($this->removeTempDir) {
			$this->log('Временная диретктория `'.$this->tempDir.'` удалена.');
			$this->_removeDir($this->tempDir);
		}
		
		$this->log('Обновление `'.$this->latestVersion.'` успешно установлено.');
		
		return true;
	}
	
	
	/*
	 * Update to the latest version
	 */
	public function update() {
		//Check for latest version
		if ((is_null($this->latestVersion)) or (is_null($this->latestUpdate))) {
			$this->checkUpdate();
		}
		
		if ((is_null($this->latestVersion)) or (is_null($this->latestUpdate))) {
			return false;
		}
		
		//Update
		if ($this->latestVersion > $this->currentVersion) {
			$this->log('Обновление...');
			
			//Add slash at the end of the path
			if ($this->tempDir[strlen($this->tempDir)-1] != '/');
				$this->tempDir = $this->tempDir.'/';
			
			if ((!is_dir($this->tempDir)) and (!mkdir($this->tempDir, 0777, true))) {
				$this->log('Временная директория `'.$this->tempDir.'` не существует и не может быть создана');
				return false;
			}
			
			if (!is_writable($this->tempDir)) {
				$this->log('Временная директория `'.$this->tempDir.'` недоступна для записи.');
				return false;
			}
			
			$updateFile = $this->tempDir.'/'.$this->latestVersion.'.zip';
			//$updateUrl = $this->updateUrl.'/'.$this->latestVersion.'.zip';
			$updateUrl = $this->latestUpdate;
			
			//Download update
			if (!is_file($updateFile)) {
				if (!$this->downloadUpdate($updateUrl, $updateFile)) {
					$this->log('Не удалось скачать обновление');
					return false;
				}
				
				$this->log('Обновление успешно скачано в файл `'.$updateFile.'`.');
			}
			else {
				$this->log('Последние обновления уже скачаны ранее `'.$updateFile.'`.');
			}
			
			//Unzip
			return $this->install($updateFile);
		}
		else {
			$this->log('Новых обновлений нет');
			return false;
		}
	}
}