<?php

require_once 'core/fiscalPrinter.php';
require_once 'fp_atol10_native.php';
require_once 'fp_atol10_errors.php';
require_once 'fp_atol10_utils.php';

class fiscalprinter_atol10 extends FiscalPrinter
{
    private $receiptType;
    private $vatBunchBeforeCloseCheck = null;
    private $moneyProfitBeforeCloseCheck = 0;

    private $FFDVersion = null; // Версия ФФД: 4 - 1.2, 3 - 1.1, 2 - 1.05, '-' - 1.0
    const FFD_VERSION_1_2 = 4;
    const FFD_VERSION_1_1 = 3;
    const FFD_VERSION_1_0_5 = 2;

    /** @var Config $config */
    private $config = null;

    // секция значений дополнительного конфига по умолчанию
    const PRINT_TAX_NUMBER_NO_DEFAULT = 0;
    const ENCASHMENT_AFTER_SHIFT_DEFAULT = 1;

    // секция названий параметров
    const PRINT_TAX_NUMBER_NO = 'print_tax_number_no';
    const ENCASHMENT_AFTER_SHIFT = 'encashment_after_shift';

    const CONFIG_DISABLE_CUT = 'disable_cut';
    const CONFIG_DISABLE_FEED = 'disable_feed';
    const CONFIG_PAPER_SALE_MODE = 'paper_sale_mode';
    const CONFIG_PAPER_SALE_RET_MODE = 'paper_sale_ret_mode';
    const CONFIG_CHECK_EMAIL = 'check_email';

    const SBP_PAYMENT_TYPE = 5; // Номер типа оплаты для системы быстрых платежей = LIBFPTR_PT_6


    // Маппинги констант бриджа в ИМЕНА констант 10го драйвера
    /** @var array VAT_MAP Маппинг НДС */
    const VAT_MAP = [
        0 => FPAtol10Native::LIBFPTR_TAX_VAT0,
        10 => FPAtol10Native::LIBFPTR_TAX_VAT10,
        18 => FPAtol10Native::LIBFPTR_TAX_VAT20, // 18 подменяем на 20
        20 => FPAtol10Native::LIBFPTR_TAX_VAT20,
        -1 => FPAtol10Native::LIBFPTR_TAX_NO,
    ];

    /** @var array ALIGNMENT_MAP Маппинг выравнивания текста */
    const ALIGNMENT_MAP = [
        0 => FPAtol10Native::LIBFPTR_ALIGNMENT_CENTER,
        1 => FPAtol10Native::LIBFPTR_ALIGNMENT_LEFT,
        2 => FPAtol10Native::LIBFPTR_ALIGNMENT_RIGHT,
    ];

    /** @var array BAUD_MAP Маппинг скоростей COM порта */
    const BAUD_MAP = [
        4 => FPAtol10Native::LIBFPTR_PORT_BR_1200,
        5 => FPAtol10Native::LIBFPTR_PORT_BR_2400,
        6 => FPAtol10Native::LIBFPTR_PORT_BR_4800,
        7 => FPAtol10Native::LIBFPTR_PORT_BR_9600,
        9 => FPAtol10Native::LIBFPTR_PORT_BR_19200,
        10 => FPAtol10Native::LIBFPTR_PORT_BR_38400,
        12 => FPAtol10Native::LIBFPTR_PORT_BR_57600,
        13 => FPAtol10Native::LIBFPTR_PORT_BR_115200,
        14 => FPAtol10Native::LIBFPTR_PORT_BR_230400,
        15 => FPAtol10Native::LIBFPTR_PORT_BR_460800,
        16 => FPAtol10Native::LIBFPTR_PORT_BR_921600,
    ];

    /** @var array RECEIPT_TYPE_MAP Маппинг типа чека */
    const RECEIPT_TYPE_MAP = [
        1 => FPAtol10Native::LIBFPTR_RT_SELL,
        2 => FPAtol10Native::LIBFPTR_RT_SELL_RETURN,
//      3 – аннулирование продажи;
        4 => FPAtol10Native::LIBFPTR_RT_BUY,
        5 => FPAtol10Native::LIBFPTR_RT_BUY_RETURN,
//      6 – аннулирование покупки
    ];

    /** @var array TAX_SYSTEM_MAP Маппинг систем налогообложения */
    const TAX_SYSTEM_MAP = [
        1 => FPAtol10Native::LIBFPTR_TT_OSN,
        2 => FPAtol10Native::LIBFPTR_TT_USN_INCOME,
        4 => FPAtol10Native::LIBFPTR_TT_USN_INCOME_OUTCOME,
        8 => FPAtol10Native::LIBFPTR_TT_ENVD,
        16 => FPAtol10Native::LIBFPTR_TT_ESN,
        32 => FPAtol10Native::LIBFPTR_TT_PATENT,
    ];

    /** @var array BARCODE_TYPE_MAP Маппинг типов штрихкодов */
    const BARCODE_TYPE_MAP = [
        'UPC-A' => FPAtol10Native::LIBFPTR_BT_UPC_A,
//        'CODE3' => "1",
        'EAN13' => FPAtol10Native::LIBFPTR_BT_EAN_13,
        'EAN8' => FPAtol10Native::LIBFPTR_BT_EAN_8,
        'UPC-E' => FPAtol10Native::LIBFPTR_BT_UPC_E,
        'ITF' => FPAtol10Native::LIBFPTR_BT_ITF,
        'CODEBAR' => FPAtol10Native::LIBFPTR_BT_CODABAR,
        'CODE93' => FPAtol10Native::LIBFPTR_BT_CODE_93,
        'CODE128' => FPAtol10Native::LIBFPTR_BT_CODE_128,
        'PDF417' => FPAtol10Native::LIBFPTR_BT_PDF417,
//        'CODE32' => "20",
//        'EAN128 CC-A/B' => "82",
//        'EAN128 CC-C' => "83",
        'QR' => FPAtol10Native::LIBFPTR_BT_QR,
    ];

    /**
     * Beep команда
     * @return array
     */
    public function Beep()
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();

        $error = $device->beep();

        if ($error < 0) {
            return ['success' => false, 'error' => $this->_internalGetLastErrorMessage()];
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * Sound команда
     * @return array
     */
    public function Sound()
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();

        require_once dirname(__FILE__) . '/fp_sound.php';
        kuznechik($device);
        $error = 0;

        if ($error < 0) {
            return ['success' => false, 'error' => $this->_internalGetLastErrorMessage()];
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * Получить инфомацию о ФР
     * @param $params
     * @return array|mixed
     */
    public function GetInfo($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

//        if (isset($params['dateTime'])) {
//            $jsonParams = [
//                'type' => 'setDateTime',
//                'dateTime' => $params['dateTime'],
//            ];
//            if ($this->isShiftClosed($device)) {
//                $this->doJsonQuery($device, $jsonParams, $res);
//            }
//        }

        $data = $this->getStatus($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $lastCheckNumber = $this->getFiscalCheckNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $inn = $this->getInn($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $machineNumber = $this->_internalGetMachineNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $result = [
            'success' => true,
            'INN' => $inn,
            'LastGlobalDocNumber' => $data['documentNumber'],
            'LastCheckNumber' => $lastCheckNumber,
            'EKLZExist' => $data['isFNPresent'],
            'ECRMode' => $data['mode'],
            'ECRModeDesc' => 'Неизвестный режим',
            'ECRMode8Status' => $data['receiptType'],
            'ECRModeStatus' => $data['mode'],
            'ECRAdvancedMode' => $data['submode'],
            'ECRAdvancedModeDesc' => 'Неизвестный подрежим',
            'SerialNumber' => str_pad($data['serialNumber'], 8, '0', STR_PAD_LEFT),
            'MachineNumber' => $machineNumber,
            'LastSessionNumber' => $data['shiftNumber'],
            'Date' => date('d.m.Y', strtotime($data['dateTime'])),
            'Time' => strtotime($data['dateTime']),
        ];

        $this->log(__METHOD__ . " ----- end. OK. " . json_encode($result, JSON_UNESCAPED_UNICODE) . " \n", 3);

        return $result;
    }

    /**
     * @return bool
     */
    public function ContinuePrintSupport()
    {
        return false;
    }

    /**
     * @param $lines
     */
    protected function beforeCheckPrintArray($lines)
    {
        $device = $this->get_NativeDevice();
        $isShiftClosed = $this->isShiftClosed($device);
        $hasOpenCheck = false;
        $operatorName = null;
        foreach ($lines as $line) {
            if ($line['action'] === 'opencheck') {
                $hasOpenCheck = true;
                if (isset($line['params'], $line['params']['operatorName'])) {
                    $operatorName = $line['params']['operatorName'];
                }
                break;
            }
        }
        if ($isShiftClosed && $hasOpenCheck) {
            $this->log(__METHOD__ . " ----- open shift in beforeCheckPrintArray\n", 3);
            if (isset($operatorName) && !empty($operatorName)) {
                $this->setOperatorName($device, $operatorName);
            }
            $device->openShift();
            $feedParams['linesQuantity'] = 5;
            $this->_do_feed($feedParams);
            $this->_do_cut();
        }
    }

    /**
     * @param $isLegacy
     * @return bool
     */
    protected function isCheckPrintArrayConvertingRequired($isLegacy)
    {
        return true;
    }

    /**
     * Заменяет все команды discount на команды print, а сумму скидки из убранных команд discount добавляет
     * в miscDiscount
     * @param $lines
     * @return array
     */
    protected function convertCheckPrintArray($lines)
    {
        return $this->finalDiscountConvertCheckPrint($lines);
    }

    /**
     * Метод печатает сменный отчет
     * @params['withcleaning'] bool - если true, то отчет с гашением
     * @param $params
     * @return array|bool|mixed
     */
    public function PrintReport($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        if (!isset($params['withcleaning'])) {
            $this->log(
                __METHOD__ . " ----- end. Fail. Не передан обязательный параметр [withcleaning]\n",
                1
            );
            return [
                'success' => false,
                'error' => 'Не передан обязательный параметр [withcleaning]',
            ];
        }
        $clean = $params['withcleaning'] != 0;

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $totalSalesCheckCount = $this->getFiscalCheckNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $status = $this->getStatus($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        $serial = str_pad($status['serialNumber'], 8, '0', STR_PAD_LEFT);
        $datetime = $status['dateTime'];
        $docNumber = $status['documentNumber'];
        $isReceiptOpen = $status['receiptType'] !== $device->LIBFPTR_RT_CLOSED;

        // проверка того может ли принтер напечатать очередной чек (игнорируем блок 24 часа)
        $result = $this->ReadyForNewCheck(true);
        if (!$result['success']) {
            $this->log(
                __METHOD__ . " ----- end. Result = fail. Принтер не готов к печати нового чека. result=" .
                json_encode($result, JSON_UNESCAPED_UNICODE) . "\n",
                1
            );
            return $result;
        }

        if ($isReceiptOpen) {
            $device->cancelReceipt();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        if ($clean) {
            $this->log(__METHOD__ . " Запуск снятия отчета с гашением\n", 3);

            if (isset($params['operatorName']) && !empty($params['operatorName'])) {
                $this->setOperatorName($device, $params['operatorName']);
            }
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }

            $device->setParam($device->LIBFPTR_PARAM_REPORT_TYPE, $device->LIBFPTR_RT_CLOSE_SHIFT);
        } else {
            $this->log(__METHOD__ . " Запуск снятия отчета без гашения\n", 3);

            $device->setParam($device->LIBFPTR_PARAM_REPORT_TYPE, $device->LIBFPTR_RT_X);
        }
        $device->report();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $this->_waitWhilePrinting();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        if (!$this->doJsonQuery($device, ['type' => 'getShiftTotals'], $res)) {
            return $res;
        }
        $totalSalesAmount = $res['shiftTotals']['receipts']['sell']['sum'];
        $session = $res['shiftTotals']['shiftNumber'];

        $machineNumber = $this->_internalGetMachineNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $result = [
            'success' => true,
            'globalchecknum' => $docNumber,
            'salesamount' => $totalSalesAmount,
            'saleschecks' => $totalSalesCheckCount,
            'session' => $session,
            'serial' => $serial,
            'MachineNumber' => $machineNumber,
            'datetime' => $datetime
        ];

        if (!$clean) {
            $result['xreportNumber'] = 0; // в атол такого регистра нет
        }

        $this->log(__METHOD__ . " ----- end. OK. " . json_encode($result, JSON_UNESCAPED_UNICODE) . " \n", 3);

        return $result;
    }

    /**
     * Возвращает дату и время ККТ
     * @param FPAtol10Native $device
     * @return mixed|string|null
     */
    protected function _internalGetDateTime($device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_DATE_TIME);
        $device->queryData();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        return (string)$device->getParamDateTime($device->LIBFPTR_PARAM_DATE_TIME);
    }

    /**
     * Возвращает серийный номер ККТ
     * @param FPAtol10Native $device
     * @return mixed|string|null
     */
    protected function _internalGetSerialNumber($device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $data = $this->getStatus($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        return str_pad($data['serialNumber'], 8, '0', STR_PAD_LEFT);
    }

    /**
     * Возвращает регистрационный номер фискального накопителя
     * @param FPAtol10Native $device
     * @return mixed|string|null
     */
    protected function _internalGetMachineNumber($device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $device->setParam($device->LIBFPTR_PARAM_FN_DATA_TYPE, $device->LIBFPTR_FNDT_TAG_VALUE);
        $device->setParam($device->LIBFPTR_PARAM_TAG_NUMBER, 1037);
        $device->fnQueryData();
        $result = $device->getParamString($device->LIBFPTR_PARAM_TAG_VALUE);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        return trim($result);
    }

    /**
     * Получает информацию о смене
     */
    public function GetShiftInfo()
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        if (!$this->doJsonQuery($device, ['type' => 'getShiftTotals'], $result)) {
            return $result;
        }

        $sellReceipts = $result['shiftTotals']['receipts']['sell'];
        $electronically = floatval($sellReceipts['payments']['electronically']);
        $sbp = floatval($sellReceipts['payments']['userPaymentType-' . self::SBP_PAYMENT_TYPE]);
        $registers['REG_OPER_SALE_COUNTS'] = $sellReceipts['count'];
        $registers['REG_CASH_SALE_SUM'] = $sellReceipts['sum'];
        $registers['REG_CASH_SALE_SUM_CASH'] = $sellReceipts['payments']['cash'];
        $registers['REG_CASH_SALE_SUM_CREDIT'] = $sellReceipts['payments']['credit'];
        $registers['REG_CASH_SALE_SUM_TARE'] = 0; // TODO найти соответствие в 10 драйвере
        $registers['REG_CASH_SALE_SUM_CARD'] = $electronically - $sbp;
        $registers['REG_CASH_SALE_SUM_SBP'] = $sbp;
        $registers['REG_CASH_SALE_SUM_COUPON'] = 0; // $sellReceipts['payments']['prepaid'];

        $buyReceipts = $result['shiftTotals']['receipts']['buy'];
        $electronically = floatval($buyReceipts['payments']['electronically']);
        $sbp = floatval($buyReceipts['payments']['userPaymentType-' . self::SBP_PAYMENT_TYPE]);
        $registers['REG_OPER_BUY_COUNTS'] = $buyReceipts['count'];
        $registers['REG_CASH_BUY_SUM'] = $buyReceipts['sum'];
        $registers['REG_CASH_BUY_SUM_CASH'] = $buyReceipts['payments']['cash'];
        $registers['REG_CASH_BUY_SUM_CREDIT'] = $buyReceipts['payments']['credit'];
        $registers['REG_CASH_BUY_SUM_TARE'] = 0; // TODO найти соответствие в 10 драйвере
        $registers['REG_CASH_BUY_SUM_CARD'] = $electronically - $sbp;
        $registers['REG_CASH_BUY_SUM_SBP'] = $sbp;

        $sellReturnReceipts = $result['shiftTotals']['receipts']['sellReturn'];
        $electronically = floatval($sellReturnReceipts['payments']['electronically']);
        $sbp = floatval($sellReturnReceipts['payments']['userPaymentType-' . self::SBP_PAYMENT_TYPE]);
        $registers['REG_OPER_SALERET_COUNTS'] = $sellReturnReceipts['count'];
        $registers['REG_CASH_SALERET_SUM'] = $sellReturnReceipts['sum'];
        $registers['REG_CASH_SALERET_SUM_CASH'] = $sellReturnReceipts['payments']['cash'];
        $registers['REG_CASH_SALERET_SUM_CREDIT'] = $sellReturnReceipts['payments']['credit'];
        $registers['REG_CASH_SALERET_SUM_TARE'] = 0; // TODO найти соответствие в 10 драйвере
        $registers['REG_CASH_SALERET_SUM_CARD'] = $electronically - $sbp;
        $registers['REG_CASH_SALERET_SUM_SBP'] = $sbp;

        $buyReturnReceipts = $result['shiftTotals']['receipts']['buyReturn'];
        $electronically = floatval($buyReturnReceipts['payments']['electronically']);
        $sbp = floatval($buyReturnReceipts['payments']['userPaymentType-' . self::SBP_PAYMENT_TYPE]);
        $registers['REG_OPER_BUYRET_COUNTS'] = $buyReturnReceipts['count'];
        $registers['REG_CASH_BUYRET_SUM'] = $buyReturnReceipts['sum'];
        $registers['REG_CASH_BUYRET_SUM_CASH'] = $buyReturnReceipts['payments']['cash'];
        $registers['REG_CASH_BUYRET_SUM_CREDIT'] = $buyReturnReceipts['payments']['credit'];
        $registers['REG_CASH_BUYRET_SUM_TARE'] = 0; // TODO найти соответствие в 10 драйвере
        $registers['REG_CASH_BUYRET_SUM_CARD'] = $electronically - $sbp;
        $registers['REG_CASH_BUYRET_SUM_SBP'] = $sbp;

        $income = $result['shiftTotals']['income'];
        $registers['REG_OPER_MONEYINPUT_COUNT'] = $income['count'];
        $registers['REG_CASH_MONEYINPUT_SUM'] = $income['sum'];

        $outcome = $result['shiftTotals']['outcome'];
        $registers['REG_OPER_MONEYOUTPUT_COUNT'] = $outcome['count'];
        $registers['REG_CASH_MONEYOUTPUT_SUM'] = $outcome['sum'];

        $registers['REG_OPER_ANNULATE_COUNT_GLOBAL'] = -1;
        $registers['REG_OPER_ANNULATE_SALE_COUNT'] = -1;
        $registers['REG_CASH_ANNULATE_SALE_SUM'] = -1;
        $registers['REG_OPER_ANNULATE_BUY_COUNT'] = -1;
        $registers['REG_CASH_ANNULATE_BUY_SUM'] = -1;
        $registers['REG_OPER_ANNULATE_SALERET_COUNT'] = -1;
        $registers['REG_CASH_ANNULATE_SALERET_SUM'] = -1;
        $registers['REG_OPER_ANNULATE_BUYRET_COUNT'] = -1;
        $registers['REG_CASH_ANNULATE_BUYRET_SUM'] = -1;

        $registers['REG_CASH_ALLCASH_SUM'] = $result['shiftTotals']['cashDrawer']['sum'];

        $registers['REG_OPER_SALE_COUNTS_DISCOUNT'] = "0";
        $registers['REG_CASH_SALE_SUM_DISCOUNT'] = "0";

        if (!$this->doJsonQuery($device, ['type' => 'getOverallTotals'], $result)) {
            return $result;
        }

        $registers['REG_CASH_TOTAL_SALE_SUM'] = $result['overallTotals']['receipts']['sell']['sum'];

        if (!$this->doJsonQuery($device, ['type' => 'getShiftStatus'], $result)) {
            return $result;
        }
        $isShiftOpened = $result['shiftStatus']['state'] !== 'closed';
        $shiftNumber = $result['shiftStatus']['number'];

        $inn = $this->getInn($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $lastDocumentNumber = $this->getLastDocumentNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            $this->log(__METHOD__ . "OpenDocumentNumber\n", 1);
            $lastDocumentNumber = 0;
        }

        $result = [
            'success' => true,
            'IsFMSessionOpen' => $isShiftOpened,
            'IsFM24HoursOver' => false,
            'SessionNumber' => $shiftNumber,
            'INN' => $inn,
            'registers' => $registers,
            'OpenDocumentNumber' => $lastDocumentNumber,
        ];

        $this->log(
            __METHOD__ . " ----- end. OK. " . json_encode($result, JSON_UNESCAPED_UNICODE) . " \n",
            3
        );

        return $result;
    }

    /**
     * @param FPAtol10Native $device
     * @param array $params
     * @param array $result
     * @return mixed
     */
    private function doJsonQuery($device, $params, &$result)
    {
        $device->setParam($device->LIBFPTR_PARAM_JSON_DATA, json_encode($params, JSON_UNESCAPED_UNICODE));
        $device->processJson();
        $resultJson = $device->getParamString($device->LIBFPTR_PARAM_JSON_DATA);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            $result = $errResult;
            return false;
        }
        $result = json_decode($resultJson, true);
        return true;
    }

    /**
     * Возвращает success=true в том случае, если принтер готов напечатать новый чек
     * @param bool $ignoreSession
     * @return array|mixed
     */
    public function ReadyForNewCheck($ignoreSession = false)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $receiptState = $this->getReceiptState($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        if ($receiptState['receiptType'] !== $device->LIBFPTR_RT_CLOSED) {
            $device->cancelReceipt();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
            $this->log(__METHOD__ . " ----- Был отменен предыдущий открытый чек\n", 3);
        }

        if (!$this->_checkPaper()) {
            $this->log(__METHOD__ . " ----- end. Fail. Нет бумаги\n", 3);
            return [
                'success' => false,
                'error' => 'Закончилась бумага. Вставьте новую ленту и повторите операцию.',
                'errcode' => EC_NO_PAPER,
                'ecrmode' => null,//$device->Mode,
                'ecradvmode' => null,//$device->AdvancedMode
            ];
        }

        if (!$ignoreSession && $this->isShiftExpired($device)) {
            $this->log(__METHOD__ . " ----- Смена истекла\n", 3);
            return $this->createSession24BlockResponse();
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * Метод регистрирует внесение денежной суммы в кассу.
     * @param $params - $params['amount'] - сумма внесения в рублях
     * @return array|mixed
     */
    public function CashIncome($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        if (!isset($params['amount']) || !is_numeric($params['amount'])) {
            $this->log(__METHOD__ . " ----- end. Fail. Не передан обязательный параметр amount.\n", 1);
            return [
                'success' => false,
                'error' => 'Не передан обязательный параметр amount', 'errcode' => EC_API_ERROR
            ];
        }
        $amount = floatval($params['amount']);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $this->printOrganizationBatch($params);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $status = $this->getStatus($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        $serial = str_pad($status['serialNumber'], 8, '0', STR_PAD_LEFT);
        $datetime = $status['dateTime'];
        $docNumber = $status['documentNumber'];
        $shiftNumber = $status['shiftNumber'];

        $device->setParam($device->LIBFPTR_PARAM_SUM, $amount);
        $device->cashIncome();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        // режем
        $this->_do_feed(['linesQuantity' => 8]);
        $this->_do_cut();

        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_CASHIN_COUNT);
        $device->queryData();
        $incomeNumber = $device->getParamInt($device->LIBFPTR_PARAM_DOCUMENTS_COUNT);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        $incomeNumberGlobal = -1;

        $this->log(__METHOD__ . " ----- end. OK. Amount: {$amount}, DocNumber: {$docNumber}\n", 3);

        return [
            'success' => true,
            'amount' => $amount,
            'globalchecknum' => $docNumber,
            'incomeNumber' => $incomeNumber,
            'incomeNumberGlobal' => $incomeNumberGlobal,
            'serial' => $serial,
            'session' => $shiftNumber,
            'datetime' => $datetime
        ];
    }

    /**
     * Метод регистрирует выплату из кассы денежной суммы
     * @param array $params - $params['amount'] - сумма выплаты в рублях
     * @return array|mixed
     */
    public function CashOutcome($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        if (!isset($params['amount']) || !is_numeric($params['amount'])) {
            $this->log(__METHOD__ . " ----- end. Fail. Не передан обязательный параметр amount.\n", 1);
            return [
                'success' => false,
                'error' => 'Не передан обязательный параметр amount', 'errcode' => EC_API_ERROR
            ];
        }
        $amount = floatval($params['amount']);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $this->printOrganizationBatch($params);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        $status = $this->getStatus($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        $serial = str_pad($status['serialNumber'], 8, '0', STR_PAD_LEFT);
        $datetime = $status['dateTime'];
        $docNumber = $status['documentNumber'];
        $shiftNumber = $status['shiftNumber'];

        $device->setParam($device->LIBFPTR_PARAM_SUM, $amount);
        $device->cashOutcome();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        // режем
        $this->_do_feed(['linesQuantity' => 8]);
        $this->_do_cut();

        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_CASHOUT_COUNT);
        $device->queryData();
        $outcomeNumber = $device->getParamInt($device->LIBFPTR_PARAM_DOCUMENTS_COUNT);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }
        $outcomeNumberGlobal = -1;

        $this->log(__METHOD__ . " ----- end. OK. Amount: {$amount}, DocNumber: {$docNumber}\n", 3);

        return [
            'success' => true,
            'amount' => $amount,
            'globalchecknum' => $docNumber,
            'outcomeNumber' => $outcomeNumber,
            'outcomeNumberGlobal' => $outcomeNumberGlobal,
            'serial' => $serial,
            'session' => $shiftNumber,
            'datetime' => $datetime,
        ];
    }

    /**
     * Функция ожидания завершения печати
     * @param int $timeoutMs Максимальное время ожидания в миллисекундах
     */
    protected function _waitWhilePrinting($timeoutMs = 10000)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);
        // про просто. Драйвер последовательный, ему нужна эта эмпирическая задержка
        usleep($timeoutMs);

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);
    }

    /**
     * Печатает строку на чеке
     * @param array $params Массив параметров
     * @return array|bool
     */
    protected function _do_print($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $str = $params['string'];

        if (is_string($str)) {
            $device = $this->get_NativeDevice();
            $device->setParam($device->LIBFPTR_PARAM_TEXT, iconv('utf-8', 'windows-1251', $str));
            $alignment = $this->getAlignment(isset($params['alignment']) ? (int)$params['alignment'] : 1);
            $device->setParam($device->LIBFPTR_PARAM_ALIGNMENT, $device->{$alignment});
            $device->setParam($device->LIBFPTR_PARAM_TEXT_WRAP, $device->LIBFPTR_TW_NONE);
            $device->printText();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }

            if ($this->_checkPaper()) {
                $this->log(__METHOD__ . " ----- end. OK.\n", 3);
                return ['success' => true];
            } else {
                $this->log(__METHOD__ . " ----- ERROR Закончилась бумага \n", 1);
                return [
                    'success' => false,
                    'error' => 'Закончилась бумага',
                    'errcode' => EC_NO_PAPER,
                    'hasStart' => true,
                ];
            }
        } else {
            $this->log(
                __METHOD__ . " ----- ERROR Неверный тип параметра для печати. Должен быть string \n",
                1
            );
            return ['success' => false, 'error' => 'Неверный тип параметра для печати. Должен быть string'];
        }
    }

    private function getAlignment($alignment)
    {
        $map = self::ALIGNMENT_MAP;
        return isset($map[$alignment]) ? $map[$alignment] : FPAtol10Native::LIBFPTR_ALIGNMENT_LEFT;
    }

    /**
     * Печатает строку на чеке с возможностью подстановки параметров
     * @param array $params
     * @return array|bool
     */
    protected function _do_printex($params)
    {
        return ['success' => true, 'info' => 'Строки нет'];
    }

    /**
     * Печатает скидку на весь чек. Эта команда не может использоваться больше одного раза в одном открытом чеке!
     * Делать скидку на позицию в протоколе АТОЛ 3.x нельзя.
     *
     * Функция на самом деле не делает скидку после 54-ФЗ.
     *
     * @param array $params Массив параметров
     * @return array|bool
     */
    protected function _do_discount($params)
    {
        return ['success' => true];
    }

    /**
     * Открывает чек для добавления позиций
     * @param array $params Массив параметров
     * @return array|bool
     */
    protected function _do_opencheck($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        if (isset($params['operatorName']) && !empty($params['operatorName'])) {
            $this->setOperatorName($device, $params['operatorName']);
        }
        if ($this->isShiftExpired($device)) {
            return $this->createSession24BlockResponse();
        }

        $this->moneyProfitBeforeCloseCheck = $this->getMoneyProfit($device);

        if (isset($params['mdlpReceiptAttr'], $params['mdlpReceiptAttr']['name'], $params['mdlpReceiptAttr']['value'])) {
            if ($this->getFFDVersion($device) < self::FFD_VERSION_1_2) {
                // Формируем отраслевой реквизит для фармы для старых версий ФФД (ниже 1.2)
                $device->setParam(1085, FPAtol10Utils::cp866($params['mdlpReceiptAttr']['name']));
                $device->setParam(1086, FPAtol10Utils::cp866($params['mdlpReceiptAttr']['value']));
                $device->utilFormTlv();
                $userAttribute = $device->getParamByteArray($device->LIBFPTR_PARAM_TAG_VALUE);
                $device->setNonPrintableParam(1084, $userAttribute);
            } else {
                // Формируем отраслевой реквизит для версий ФФД 1.2 и выше
                if ($params['mdlpReceiptAttr']['name'] === 'mdlp') {
                    // Продажа лекарств
                    $device->setParam(1262, '020'); // идентификатор ФОИВ = Министерство здравоохранения Российской Федерации
                    $device->setParam(1263, '14.12.2018'); // Дата документа основания
                    $device->setParam(1264, '1556'); // Номер документа основания
                    $device->setParam(1265,
                        'tm=mdlp&' . FPAtol10Utils::cp866($params['mdlpReceiptAttr']['value'])
                    ); // Ex: tm=mdlp&sid=00000000105200
                    $device->utilFormTlv();
                    $userAttribute = $device->getParamByteArray($device->LIBFPTR_PARAM_TAG_VALUE);
                    $device->setNonPrintableParam(1261, $userAttribute);
                }
            }
        }

        $this->receiptType = $this->getReceiptType(isset($params['checktype']) ? $params['checktype'] : 1);
        $device->setParam($device->LIBFPTR_PARAM_RECEIPT_TYPE, $device->{$this->receiptType});
        $paperSaleMode = $this->config->getParamInt(self::CONFIG_PAPER_SALE_MODE);
        $paperSaleRetMode = $this->config->getParamInt(self::CONFIG_PAPER_SALE_RET_MODE);
        $email = $this->config->getParameter(self::CONFIG_CHECK_EMAIL, null);
        $paperMode = ($this->receiptType === FPAtol10Native::LIBFPTR_RT_SELL) ? $paperSaleMode : $paperSaleRetMode;

        /************************************************************/
        /*Временное решение для смены системы налогообложения в чеке*/
        /************************************************************/
        $str = isset($params['customerContact']) && is_string($params['customerContact']) ?
            $params['customerContact'] : '';
        if ((strpos($str, '-') === 0) && (is_numeric($val = substr($str, 1)))) {
            $params['customerContact'] = '';
            $taxSystem = $this->getTaxSystem($val);
            $device->setParam(1055, $device->{$taxSystem}); // Применяемая система налогообложения (СНО)
        }

        /***************************************/
        $customerContact = null;
        if (
            isset($params['customerContact']) && !empty($params['customerContact']) &&
            (preg_match('/[^A-Za-z0-9@\-().+ _]+/', $params['customerContact']) === 0)
        ) {
            $customerContact = $params['customerContact'];
        } elseif (!empty($email)) {
            $customerContact = $email; // Подставляем электронный адрес по умолчанию
        }
        
        if (!is_null($customerContact)) {
            $device->setParam(1008, $customerContact); // Электронный адрес / телефон получателя
        }
        
        if (($paperMode === 2) || (($paperMode === 1) && !empty($customerContact))) {
            $device->setParam($device->LIBFPTR_PARAM_RECEIPT_ELECTRONICALLY, $str !== '+');
        }
        
        $device->openReceipt();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }
        $this->_waitWhilePrinting();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $this->vatBunchBeforeCloseCheck = $this->getVatBunch($device, $this->receiptType);

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * @param $receiptType
     * @return string
     */
    private function getReceiptType($receiptType)
    {
        $map = self::RECEIPT_TYPE_MAP;
        return isset($map[$receiptType]) ? $map[$receiptType] : FPAtol10Native::LIBFPTR_RT_SELL;
    }

    private function getTaxSystem($taxSystem)
    {
        $map = self::TAX_SYSTEM_MAP;
        return isset($map[$taxSystem]) ? $map[$taxSystem] : FPAtol10Native::LIBFPTR_TT_OSN;
    }

    /**
     * Закрывает чек
     *
     * @param array $params Массив параметров
     * @return array|bool
     */
    protected function _do_closecheck($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $paymentArray = $params['payment'];

        $summCash = isset($paymentArray['cash']) ? $paymentArray['cash'] : 0;
        $summCredit = isset($paymentArray['credit']) ? $paymentArray['credit'] : 0;
        $summTare = isset($paymentArray['tare']) ? $paymentArray['tare'] : 0;
        $summCard = isset($paymentArray['card']) ? $paymentArray['card'] : 0;
        $summSbp = isset($paymentArray['sbp']) ? $paymentArray['sbp'] : 0;
        $summCoupon = isset($paymentArray['coupon']) ? $paymentArray['coupon'] : 0;
        $SummBonus = isset($paymentArray['bonus']) ? $paymentArray['bonus'] : 0;
        $SummMiscDiscount = isset($paymentArray['miscDiscount']) ? $paymentArray['miscDiscount'] : 0;
        $roundAmount = isset($paymentArray['roundAmount']) ? $paymentArray['roundAmount'] : 0;

        if (!is_numeric($summCash) || !is_numeric($summCredit) || !is_numeric($summTare) || !is_numeric($summCard) ||
            !is_numeric($summSbp)  || !is_numeric($SummBonus) || !is_numeric($SummMiscDiscount))
        {
            $this->log(
                __METHOD__ . " Ошибка. Переданная сумма неверного типа " .
                "Summ1=$summCash; Summ2=$summCredit; Summ3=$summTare; Summ4=$summCard; SummSbp=$summSbp; SummBonus=$SummBonus; " .
                "SummMiscDiscount=$SummMiscDiscount; \n", 1
            );
            return [
                'success' => false,
                'error' => "Ошибка. Переданная сумма неверного типа\n",
                'errcode' => EC_API_ERROR,
                'Summ1' => $summCash,
                'Summ2' => $summCredit,
                'Summ3' => $summTare,
                'Summ4' => $summCard,
                'Summ5' => $summSbp,
                'SummBonus' => $SummBonus,
                'SummMiscDiscount' => $SummMiscDiscount,
            ];
        }

        if ($roundAmount > 0) {
            if (!isset($params['checkamount'])) {
                return [
                    'success' => false,
                    'error' => "Ошибка. Не передана сумма чека\n",
                    'errcode' => EC_API_ERROR,
                ];
            }
            $receiptSum = floatval($params['checkamount']);

            $this->log(__METHOD__ . " ----- Округление. Сумма чека: $receiptSum. Бонусы: $SummBonus.\n", 3);

            $receiptSum -= floatval($SummBonus);

            $device->setParam($device->LIBFPTR_PARAM_SUM, $receiptSum);
            $device->receiptTotal();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        if ($summCredit > 0) {
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_TYPE, $device->LIBFPTR_PT_CREDIT);
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_SUM, floatval($summCredit));
            $device->payment();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        // TODO найти сопоставление для TARE
//        if ($summTare > 0) {
//            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_TYPE, );
//            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_SUM, floatval($summTare));
//            $device->payment();
//            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
//                return $errResult;
//            }
//        }

        if ($summCard > 0) {
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_TYPE, $device->LIBFPTR_PT_ELECTRONICALLY);
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_SUM, floatval($summCard));
            $device->payment();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        if ($summSbp > 0) {
            $prop = 'LIBFPTR_PT_' . (self::SBP_PAYMENT_TYPE + 1);
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_TYPE, $device->$prop);
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_SUM, floatval($summSbp));
            $device->payment();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        if ($summCoupon > 0) {
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_TYPE, $device->LIBFPTR_PT_PREPAID);
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_SUM, floatval($summCoupon));
            $device->payment();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        if ($summCash > 0) {
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_TYPE, $device->LIBFPTR_PT_CASH);
            $device->setParam($device->LIBFPTR_PARAM_PAYMENT_SUM, floatval($summCash));
            $device->payment();
            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        $receiptState = $this->getReceiptState($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }
        $sum = $receiptState['sum'];
        $remainder = $receiptState['remainder'];
        $change = $receiptState['change'];
        $lastDocumentNumber = $receiptState['documentNumber'];

        if ($remainder > 0) {
            $this->log(__METHOD__ . " ----- ERROR Покупка оплачена не полностью\n", 1);
            return [
                'success' => false,
                'error' => 'Покупка оплачена не полностью',
                'errcode' => EC_AMOUNT_CHECK_FAIL,
                'hasStart' => true
            ];
        }

        $device->closeReceipt();
        $device->checkDocumentClosed();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        if (!$device->getParamInt($device->LIBFPTR_PARAM_DOCUMENT_CLOSED)) {
            $device->cancelReceipt();
            $this->log(__METHOD__ . " ----- ERROR Чек не закрылся и был отменен\n", 1);
            return [
                'success' => false,
                'error' => "Чек не закрылся и был отменен\n",
                'errcode' => EC_DEVICE_ERROR,
            ];
        }

        $this->_waitWhilePrinting();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $device->setParam($device->LIBFPTR_PARAM_FN_DATA_TYPE, $device->LIBFPTR_FNDT_LAST_DOCUMENT);
        $device->fnQueryData();
        $fiscalSign = $device->getParamString($device->LIBFPTR_PARAM_FISCAL_SIGN);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }
        $this->fiscalSign = $fiscalSign;

        $lastCheckNumber = $this->getFiscalCheckNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }
        $this->_lastCheckNumber = $lastCheckNumber;

        $this->_lastCheckAmount = $sum;
        $this->_change = $change;
        $this->_lastGlobalCheckNumber = $lastDocumentNumber;

        $sessionNumber = $this->getShiftNumber($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }
        $this->_sessionNumber = $sessionNumber;

        $vatBunchAfterClose = $this->getVatBunch($device, $this->receiptType);
        $moneyProfitAfterClose = $this->getMoneyProfit($device);

        $vatBunchDelta = $this->getVatBunchDelta($this->vatBunchBeforeCloseCheck, $vatBunchAfterClose);
        $this->vat10Sum = $vatBunchDelta['vat10'];
        $this->vat18Sum = $vatBunchDelta['vat18'];
        $this->vat20Sum = $vatBunchDelta['vat20'];

        $this->isMoneySumChanged = $this->moneyProfitBeforeCloseCheck !== $moneyProfitAfterClose;

        if (!$this->_checkPaper()) {
            $this->log(__METHOD__ . " ----- ERROR Закончилась бумага \n", 1);
            return ['success' => false, 'error' => 'Закончилась бумага', 'errcode' => EC_NO_PAPER, 'hasStart' => true];
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * Проводит операцию продажа, печатает ее на чеке
     * @param array $params Массив параметров
     * @return array
     */
    protected function _do_sale($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
//            if ($errResult['device_errnum'] != -12) {
                return $errResult;
//            }
        }

        $taxType = $this->getTaxType(isset($params['vat_percent']) ? $params['vat_percent'] : 0);

        $itemParams = [
            'taxType' => $taxType,
            'price' => isset($params['priceWithFinalDiscount']) ? $params['priceWithFinalDiscount'] : $params['price'],
            'nameWIN1251' => iconv('utf-8', 'windows-1251', $params['item']),
            'nameUTF8' => $params['item'],
            'quantity' => $params['quantity'],
            'department' => isset($params['department']) ? $params['department'] : 0,
            'mdlpReceiptLineAttr' => isset($params['mdlpReceiptLineAttr']) ? $params['mdlpReceiptLineAttr'] : null,
        ];

        $this->log(
            __METHOD__ .
            " - [{$itemParams['nameUTF8']}] [{$itemParams['price']}] [{$itemParams['quantity']}]\n",
            3
        );

        if (isset($params['itemCodes']) || isset($params['itemMarks'])) {
            if (isset($params['itemCodes'])) $itemParams['itemCodes'] = $params['itemCodes'];
            if (isset($params['itemMarks'])) $itemParams['itemMarks'] = $params['itemMarks'];
            $errResult = ($this->getFFDVersion($device) >= self::FFD_VERSION_1_2)
                ? $this->addMarkedItem($device, $itemParams)
                : $this->addMarkedItemLegacy($device, $itemParams);
        } else {
            $errResult = $this->addItem($device, $itemParams);
        }
        if ($errResult) {
            return $errResult;
        }

        $printTaxNumberNo = $this->config->getParameter(
            self::PRINT_TAX_NUMBER_NO,
            self::PRINT_TAX_NUMBER_NO_DEFAULT
        );
        if ($taxType === FPAtol10Native::LIBFPTR_TAX_NO && $printTaxNumberNo == 1) {
            $params['string'] = 'без НДС';
            $params['alignment'] = 2; // выравниваем по правому краю, работает только в протоколе АТОЛ 2.x
            $this->_do_print($params);
        }
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }
        usleep(10000);

        if (!$this->_checkPaper(true)) {
            $this->log(__METHOD__ . " ------ ERROR Закончилась бумага\n", 1);
            return [
                'success' => false,
                'error' => 'Закончилась бумага',
                'errcode' => EC_NO_PAPER,
                'hasStart' => true
            ];
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    private function getTaxType($vatPercent)
    {
        $map = self::VAT_MAP;
        return isset($map[$vatPercent]) ? $map[$vatPercent] : FPAtol10Native::LIBFPTR_TAX_NO;
    }

    /**
     * @param FPAtol10Native $device
     * @param $params
     * @return bool|mixed
     */
    private function addItem($device, $params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device->setParam($device->LIBFPTR_PARAM_COMMODITY_NAME, $params['nameWIN1251']);
        $device->setParam($device->LIBFPTR_PARAM_PRICE, $params['price']);
        $device->setParam($device->LIBFPTR_PARAM_QUANTITY, $params['quantity']);
        $device->setParam($device->LIBFPTR_PARAM_TAX_TYPE, $device->{$params['taxType']});
        $device->setParam($device->LIBFPTR_PARAM_TAX_SUM, 0); // ККТ посчитает автоматически
        $device->setParam($device->LIBFPTR_PARAM_TAX_MODE, $device->LIBFPTR_TM_POSITION);
        $device->setParam($device->LIBFPTR_PARAM_DEPARTMENT, $params['department']);

        $device->registration();

        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            return $errResult;
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return false;
    }

    /**
     * Проверяет марку на сервере. Возвращает FALSE если проверка не получилась
     *
     * @param FPAtol10Native $device
     * @param string $mark
     * @param float $quantity
     * @param int $status
     * @return boolean|int
     */
    private function validateItemMark($device, $mark, $quantity, $status)
    {
        $this->log(__METHOD__ . " ----- start mark validation [$mark], qty=[$quantity]\n", 3);

        // Запускаем предварительную проверку марки
        $device->setParam($device->LIBFPTR_PARAM_MARKING_CODE_TYPE, $device->LIBFPTR_MCT12_AUTO);
        $device->setParam($device->LIBFPTR_PARAM_MARKING_CODE, $mark);
        $device->setParam($device->LIBFPTR_PARAM_MARKING_CODE_STATUS, $status);
        $device->setParam($device->LIBFPTR_PARAM_QUANTITY, 1); // Количество
        $device->setParam($device->LIBFPTR_PARAM_MEASUREMENT_UNIT, $device->LIBFPTR_IU_PIECE);
        $device->setParam($device->LIBFPTR_PARAM_MARKING_WAIT_FOR_VALIDATION_RESULT, true);
        $device->setParam($device->LIBFPTR_PARAM_MARKING_PROCESSING_MODE, 0);
		if ($quantity < 1) {
			$device->setParam(
				$device->LIBFPTR_PARAM_MARKING_FRACTIONAL_QUANTITY, 
				intval($quantity * 1000) .'/1000'
			);
		}

        $device->beginMarkingCodeValidation();
        $vResult = 0;
        for($i = 0; $i < 250; $i++) { // 25 секунд макс
            $device->getMarkingCodeValidationStatus();
            $vResult = $device->getParamInt($device->LIBFPTR_PARAM_MARKING_CODE_VALIDATION_READY);
            if ($vResult) break;
            usleep(100000); // 0.1 секунда
        }
        if (!$vResult) {
            $this->log(__METHOD__ . " - [$mark] mark validation timeout \n", 1);
            $device->cancelMarkingCodeValidation();
            return false;
        }

        if ($vResult !== 15) {
            // Проверка не пройдена
            $this->log(
                __METHOD__ . " - [$mark] mark validation warning: " .
                print_r([
                    'result' => $vResult,
                    'isRequestSent' => $device->getParamInt($device->LIBFPTR_PARAM_IS_REQUEST_SENT),
                    'error' => $device->getParamInt($device->LIBFPTR_PARAM_MARKING_CODE_ONLINE_VALIDATION_ERROR),
                    'errorDescription' => iconv('windows-1251', 'utf-8', $device->getParamString($device->LIBFPTR_PARAM_MARKING_CODE_ONLINE_VALIDATION_ERROR_DESCRIPTION)),
                    'info' => $device->getParamInt(2109),
                    'processingResult' => $device->getParamInt(2005),
                    'processingCode' => $device->getParamInt(2105)
                ], true) .
                " \n",
                2
            );
        }

        return $vResult;
    }

    /**
     * Регистрация маркированного товара для старой версии ФФД (до 1.2)
     * @param FPAtol10Native $device
     * @param $params
     * @return bool|mixed
     */
    private function addMarkedItemLegacy($device, $params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $itemCodes = explode(',', $params['itemCodes']);
        $quantity = $params['quantity'];
        if (isset($params['mdlpReceiptLineAttr'])) {
            $mdlpReceiptLineAttr = FPAtol10Utils::cp866($params['mdlpReceiptLineAttr']);
        }

        for ($i = 0; $quantity > 0.0; $quantity -= 1.0, $i++) {
            $itemCode = isset($itemCodes[$i]) ? $itemCodes[$i] : '';
            $this->log(__METHOD__ . " - [$itemCode]\n", 3);

            if ($itemCode) {
                $device->setParamStrHex(1162, $itemCode);
            }

            if (isset($mdlpReceiptLineAttr)) {
                $device->setNonPrintableParam(1191, $mdlpReceiptLineAttr);
            }

            $device->setParam($device->LIBFPTR_PARAM_COMMODITY_NAME, $params['nameWIN1251']);
            $device->setParam($device->LIBFPTR_PARAM_PRICE, $params['price']);
            $unitQuantity = min($quantity, 1.0);
            $device->setParam($device->LIBFPTR_PARAM_QUANTITY, $unitQuantity);
            $device->setParam($device->LIBFPTR_PARAM_POSITION_SUM, $params['price'] * $unitQuantity);
            $device->setParam($device->LIBFPTR_PARAM_TAX_TYPE, $device->{$params['taxType']});
            $device->setParam($device->LIBFPTR_PARAM_TAX_SUM, 0); // ККТ посчитает автоматически
            $device->setParam($device->LIBFPTR_PARAM_TAX_MODE, $device->LIBFPTR_TM_POSITION);
            $device->setParam($device->LIBFPTR_PARAM_DEPARTMENT, $params['department']);

            $device->registration();

            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return false;
    }

    /**
     * Регистрация маркированного товара для версии ФФД 1.2
     * @param FPAtol10Native $device
     * @param $params
     * @return bool|mixed
     */
    private function addMarkedItem($device, $params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $itemMarks = explode(',', $params['itemMarks']);
        $quantity = $params['quantity'];
        if (isset($params['mdlpReceiptLineAttr'])) {
            $mdlpReceiptLineAttr = FPAtol10Utils::cp866($params['mdlpReceiptLineAttr']);
        }

        for ($i = 0; $quantity > 0.0; $quantity -= 1.0, $i++) {
            $mark = isset($itemMarks[$i]) ? $itemMarks[$i] : '';
            $status = 2;
            $unitQuantity = min($quantity, 1.0);
            $this->log(__METHOD__ . " - [$mark]\n", 3);

            $markValidation = null;
            if ($mark) {
                $markValidation = $this->validateItemMark($device, $mark, $unitQuantity, $status);
                if ($markValidation === false) {
                    $this->log(__METHOD__ . " - ERROR Таймаут при обращении к онлайн-сервису проверки марки [$mark] \n", 1);
                    return ['success' => false, 'error' => 'Критическая ошибка. Таймаут при обращении к онлайн-сервису проверки марки ' . $mark];
                }
                $device->acceptMarkingCode();
            }

            if (isset($mdlpReceiptLineAttr)) {
				// Продажа лекарств
				$mdlpAttrValue = isset($params['mdlpReceiptLineAttrValue']) ? $params['mdlpReceiptLineAttrValue'] : '';
				$device->setParam(1262, '020'); // идентификатор ФОИВ = Министерство здравоохранения Российской Федерации
				$device->setParam(1263, '14.12.2018'); // Дата документа основания
				$device->setParam(1264, '1556'); // Номер документа основания
				$device->setParam(1265, 'tm=mdlp&' . $mdlpAttrValue); // Ex: tm=mdlp&sid=00000000105200
				$device->utilFormTlv();
				$userAttribute = $device->getParamByteArray($device->LIBFPTR_PARAM_TAG_VALUE);
				$device->setNonPrintableParam(1260, $userAttribute);
            }
			
            $device->setParam($device->LIBFPTR_PARAM_COMMODITY_NAME, $params['nameWIN1251']);
            $device->setParam($device->LIBFPTR_PARAM_PRICE, $params['price']);
            $device->setParam($device->LIBFPTR_PARAM_QUANTITY, $unitQuantity);
            // $device->setParam($device->LIBFPTR_PARAM_MEASUREMENT_UNIT, $device->LIBFPTR_IU_PIECE);
            $device->setParam($device->LIBFPTR_PARAM_POSITION_SUM, $params['price'] * $unitQuantity);
            $device->setParam($device->LIBFPTR_PARAM_TAX_TYPE, $device->{$params['taxType']});
            $device->setParam($device->LIBFPTR_PARAM_TAX_SUM, 0); // ККТ посчитает автоматически
            $device->setParam($device->LIBFPTR_PARAM_TAX_MODE, $device->LIBFPTR_TM_POSITION);
            $device->setParam($device->LIBFPTR_PARAM_DEPARTMENT, $params['department']);
			/*if ($quantity < 1) {
				$device->setParam($device->LIBFPTR_PARAM_MARKING_FRACTIONAL_QUANTITY, intval($quantity * 1000) .'/1000');
				$device->setParam($device->LIBFPTR_PARAM_PRICE, $params['price'] * $unitQuantity);
			}*/

            if (is_null($markValidation)) {
                // Продаем товар как немаркированный
                $device->setParam(1212, 32); //Обязательный параметр для ФФД 1.2!!! 32 - товар не имеющий код маркировки, 33 - товар имеющий код маркировки
            } else {
                // Продаем товар с маркировкой
                $device->setParam(1212, 33); //Обязательный параметр для ФФД 1.2!!! 32 - товар не имеющий код маркировки, 33 - товар имеющий код маркировки
                $device->setParam($device->LIBFPTR_PARAM_MARKING_CODE, $mark);
                $device->setParam($device->LIBFPTR_PARAM_MARKING_CODE_STATUS, $status);
                $device->setParam($device->LIBFPTR_PARAM_MARKING_CODE_ONLINE_VALIDATION_RESULT, $markValidation);
                $device->setParam($device->LIBFPTR_PARAM_MARKING_PROCESSING_MODE, 0);
            }

            $device->registration();

            if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
                return $errResult;
            }
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return false;
    }

    /**
     * Проводит операцию возврат продажи
     * @param array $params Массив параметров
     * @return array
     */
    protected function _do_returnsale($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        return $this->_do_sale($params);
    }

    /**
     * Возвращает true если есть бумага
     * @param bool $fast
     * @return bool
     */
    protected function _checkPaper($fast = false)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();

        $this->log(__METHOD__ . " Update Status \n", 3);

        $data = $fast ? $this->getShortStatus($device) : $this->getStatus($device);
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return false;
        }
        $isPrinterOverheat = isset($data['isPrinterOverheat']) ? $data['isPrinterOverheat'] : false;

        $this->log(__METHOD__ . " Status updated\n", 3);
        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        // если есть бумага и печатающая головка не перегрелась
        return $data['isPaperPresent'] && !$isPrinterOverheat;
    }

    /**
     * Инициализация ККМ
     */
    protected function _initialize() {
        parent::_initialize();

        $this->config = Config::getInstance();
        if (!isset($this->_nativeDevice)) {
            $this->_nativeDevice = new FPAtol10Native(
                FPAtol10Native::OLE_FISCAL_PRINTER_ATOL_DRIVER_10_PROGID
            );
        }
    }

    /**
     * @return FPAtol10Native
     */
    protected function &get_NativeDevice()
    {
        return parent::get_NativeDevice();
    }

    /**
     * Открытие порта
     * @param FPAtol10Native $device объект драйвера
     * @return bool true - порт удачно открыт
     * @throws Exception исключение в случае каких-то сбоев с открытием порта
     */
    protected function _internalOpen($device = null)
    {
        $this->log(__METHOD__ . " ----- start. Connecting to driver...\n", 3);

        $device = isset($device) ? $device : $this->get_NativeDevice();

//        if ($this->_deviceModel != -1) {
//            $device->Model = $this->_deviceModel;
//        }
        $device->setSingleSetting($device->LIBFPTR_SETTING_MODEL, $device->LIBFPTR_MODEL_ATOL_AUTO);

        $this->log(__METHOD__ . " Setup device COM Port...\n", 3);
        if ($this->_deviceCOMPortNumber < 0 && $this->_deviceIpAddress < 0) {
            $this->log(__METHOD__ . " ----- ERROR... want to specify a port (\$_GET['com']) \n", 1);
            throw new Exception("You have to specify COM or IP\n");
        }

        $this->log(__METHOD__ . " Detect COM Port. Setup device COM Port at {$this->_deviceCOMPortNumber} \n", 3);

        if ($this->_deviceIpAddress > 0) {
            throw new Exception("Not available in this version\n");
        } else {
            $device->setSingleSetting($device->LIBFPTR_SETTING_PORT, $device->LIBFPTR_PORT_COM);
            $device->setSingleSetting($device->LIBFPTR_SETTING_COM_FILE, $this->_deviceCOMPortNumber);
        }

        if ($this->_deviceCOMPortBaud > 0) {
            $this->log(__METHOD__ . " ----- com port baud config detected. Setup baud \n", 3);

            if ($this->_deviceCOMPortBaud > 13) {
                $this->log(__METHOD__ . " ----- ERROR baudRate index too match. It must 1-13 \$_GET['com'] \n", 1);
                throw new Exception("ERROR baudRate index too match. It must 1-13 \n");
            }

            $baudRate = $this->getBaudRate($this->_deviceCOMPortBaud);
            $device->setSingleSetting($device->LIBFPTR_SETTING_BAUDRATE, $device->{$baudRate});
        }

        $error = $device->applySingleSettings();

        if ($error < 0) {
            $errorCode = $this->_internalGetLastError($device);
            $errorDescription = $this->_internalGetLastErrorMessage($device);
            $this->log(__METHOD__ . " ----- ERROR Device initializing failed. ResultCode: $errorCode [$errorDescription] \n", 1);
        }

        $error = $device->open();

        if ($error < 0) {
            $errorCode = $this->_internalGetLastError($device);
            $errorDescription = $this->_internalGetLastErrorMessage($device);
            $this->log(
                __METHOD__ .
                " ----- ERROR Device opening failed. ResultCode: $errorCode [$errorDescription]\n",
                1
            );
        }

        $comPort = $device->getSingleSetting($device->LIBFPTR_SETTING_COM_FILE);
        $baudRate = $device->getSingleSetting($device->LIBFPTR_SETTING_BAUDRATE);
        $this->log(__METHOD__ . " ----- enable device \n", 3);
        $this->log(
            __METHOD__ .
            " Using COM-port: $comPort:$baudRate " .
            "(settings: {$this->_deviceCOMPortNumber}:{$this->_deviceCOMPortBaud} model:{$this->_deviceModel})...\n",
                   3
        );

        $this->log(__METHOD__ . " ----- check is device opened and ResultCode \n", 3);
        if (!$device->isOpened()) {
            $errorCode = $this->_internalGetLastError($device);
            $errorDescription = $this->_internalGetLastErrorMessage($device);
            $this->log(
                __METHOD__ .
                " ----- ERROR COM Port not open. ResultCode: $errorCode [$errorDescription]\n",
                1
            );
            throw new Exception("COM Port not open. ResultCode: $errorCode\n");
        }

        $this->log(__METHOD__ . " ----- Connected. OK.\n", 3);

        return true;
    }

    /**
     * @param $baudRate
     * @return string
     * @throws Exception
     */
    private function getBaudRate($baudRate)
    {
        $map = self::BAUD_MAP;
        if (isset($map[$baudRate])) {
            return $map[$baudRate];
        }
        $this->log(__METHOD__ . " ----- ERROR baudRate index $baudRate is unsupported\n", 1);
        throw new Exception("ERROR baudRate index $baudRate is unsupported\n");
    }

    /**
     * Закрытие порта
     * @param FPAtol10Native $deviceInput объект драйвера
     * @return bool true - порт удачно закрыт
     */
    protected function _internalClose($deviceInput = null)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = isset($deviceInput) ? $deviceInput : $this->get_NativeDevice();

        $this->log(__METHOD__ . " Disconnecting from device...\n", 3);

        $result = $device->close();

        if ($result < 0) {
            return false;
        }

        $this->_connected = false;

        $this->log(__METHOD__ . " ----- end. Disonnected. OK.\n", 3);

        return true;
    }

    /**
     * Получить номер последней ошибки
     * @param FPAtol10Native $device
     * @return int
     */
    protected function _internalGetLastError($device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        return $device->errorCode();
    }

    /**
     * Описание последней ошибки
     * @param FPAtol10Native $device
     * @return string конвертированная в utf-8 строка ошибки
     */
    protected function _internalGetLastErrorMessage($device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        return FPAtol10Utils::convertToString($device->errorDescription());
    }

    /**
     * Инициализация конфига
     * @param $params
     * @return array|mixed
     */
    public function InitConfig($params)
    {
        $this->log(__METHOD__ . " ----- start init config\n", 3);

        $device = $this->get_NativeDevice();
        if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) {
            return $errResult;
        }

        if (!$this->isShiftClosed($device)) {
            $this->log(__METHOD__ . " ----- end. Skip. OK.\n", 3);

            return ['success' => true, 'info' => 'configure skipped'];
        }

        $encashmentAfterShift = $this->config->getParameter(
            self::ENCASHMENT_AFTER_SHIFT,
            self::ENCASHMENT_AFTER_SHIFT_DEFAULT
        );

        $jsonParams = [
            'type' => 'setDeviceParameters',
            'deviceParameters' => [
                // Количество печатаемых строк после чека на ЧЛ
                (object)[
                    'key' => 14,
                    'value' => 0,
                ],
                // Печатать на ЧЛ сообщение о вводе даты-времени
                (object)[
                    'key' => 31,
                    'value' => 0, // не печатать
                ],
                // Количество строк заголовка клише
                (object)[
                    'key' => 35,
                    'value' => 0,
                ],
                // Включим звуковую индикацию при возникновении ошибок
                (object)[
                    'key' => 37,
                    'value' => 1,
                ],
                // Включим инкассацию при закрытии смены
                (object)[
                    'key' => 4,
                    'value' => (int)$encashmentAfterShift,
                ],
            ],
        ];
        if (!$this->doJsonQuery($device, $jsonParams, $result)) {
            return $result;
        }

        $this->log(
            __METHOD__ . " ----- Settings applied: " .
            json_encode($result, JSON_UNESCAPED_UNICODE) . " \n",
            3
        );

        // производим настройку времени ФР
//        $jsonParams = [
//            'type' => 'setDateTime',
//            'dateTime' => $params['dateTime'],
//        ];
//        if (!$this->doJsonQuery($device, $jsonParams, $result)) {
//            return $result;
//        }

//        $this->log(__METHOD__ . " ----- Datetime set to: {$params['dateTime']}\n", 3);
        $this->log(__METHOD__ . " ----- end. Configured. OK.\n", 3);

        return ['success' => true, 'info' => 'configure complete'];
    }

    /**
     * Обрезка чека
     *
     * @param array $params Массив параметров. Необязательный параметр $params[timeout] - время в миллисекундах,
     * сколько ждать принтер чеков
     * @return array|bool
     */
    protected function _do_cut($params = [])
    {
        if ($this->config->getParamBool(self::CONFIG_DISABLE_CUT, false)) {
            return ['success' => true, 'warning' => 'Cut function disabled in bridge config file'];
        }

        $this->log(
            __METHOD__ .
            " ----- start params: [" . json_encode($params, JSON_UNESCAPED_UNICODE) . "]\n",
            3
        );

        $device = $this->get_NativeDevice();

        // если параметр не передан, то по умолчанию полная отрезка
        $params['fullcut'] = !isset($params['fullcut']) ? false : $params['fullcut'];

        try {
            if ($params['fullcut']) {
                $device->setParam($device->LIBFPTR_PARAM_CUT_TYPE, $device->LIBFPTR_CT_FULL);
            } else {
                $device->setParam($device->LIBFPTR_PARAM_CUT_TYPE, $device->LIBFPTR_CT_PART);
            }
            $device->cut();
        } catch (Exception $e) {
            //Skip cut errors
            $this->log(
                __METHOD__ . " Cut method not supported (or just error occured) {$e->getMessage()} \n",
                2
            );
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * Прогон чека на несколько строк вперед
     *
     * @param array $params Массив параметров
     * @return array|bool
     */
    protected function _do_feed($params)
    {
        if ($this->config->getParamBool(self::CONFIG_DISABLE_FEED, false)) {
            return ['success' => true, 'warning' => 'Feed function disabled in bridge config file'];
        }
        $this->log(__METHOD__ . " ----- start\n", 3);

        $quantity = $params['linesQuantity'];

        for ($i = 0; $i < $quantity; ++$i) {
            $this->_do_print(['string' => ' ']);
        }

        $this->_waitWhilePrinting();

        if (!$this->_checkPaper()) {
            $this->log(__METHOD__ . " ----- ERROR Закончилась бумага \n", 1);
            return ['success' => false, 'error' => 'Закончилась бумага', 'errcode' => EC_NO_PAPER, 'hasStart' => true];
        }

        $this->log(__METHOD__ . " ----- end. OK.\n", 3);

        return ['success' => true];
    }

    /**
     * Открыть денежный ящик
     * @param array $params
     * @return array
     */
    protected function _do_openDrawer($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);
//        $device = $this->get_NativeDevice();
        //if ($errResult = $this->_isDeviceFailed(__METHOD__, $device)) return $errResult;
//        $device->setParam($device->LIBFPTR_PARAM_DRAWER_OFF_TIMEOUT, 50);
//        $device->setParam($device->LIBFPTR_PARAM_DRAWER_ON_QUANTITY, 3);
//        $device->setParam($device->LIBFPTR_PARAM_DRAWER_OFF_TIMEOUT, 50);
//        $device->openDrawer();
        $this->log(__METHOD__ . " ----- end. OK\n", 3);
        return ['success' => true];
    }

    /**
     * Печать штрихкода ean13
     * @param array $params Массив параметров
     * @return array|bool
     */
    protected function _do_printbarcode($params)
    {
        $this->log(__METHOD__ . " ----- start\n", 3);

        $device = $this->get_NativeDevice();

        $type = isset($params['type']) ? $params['type'] : 'EAN13';
        if (!array_key_exists($type, self::BARCODE_TYPE_MAP)) {
            $this->log(__METHOD__ . " ----- ERROR Тип штрих-кода не поддерживается $type \n", 1);
            return ['success' => false, 'error' => 'Тип штрих-кода не поддерживается', 'errcode' => EC_API_ERROR];
        }

        $value = iconv('utf-8', 'windows-1251', $params['barcode']);
        $device->setParam($device->LIBFPTR_PARAM_BARCODE, $value);
        $device->setParam($device->LIBFPTR_PARAM_BARCODE_TYPE, $device->{self::BARCODE_TYPE_MAP[$type]});
        $device->setParam($device->LIBFPTR_PARAM_ALIGNMENT, $device->LIBFPTR_ALIGNMENT_CENTER);
        $scale = 1;
        if ($type === 'QR') {
            $scale = 6;
            if (isset($params['scale']) && $params['scale'] < 10) {
                $scale = $params['scale'];
            }
        }
        $device->setParam($device->LIBFPTR_PARAM_SCALE, $scale);
        $device->setParam($device->LIBFPTR_PARAM_BARCODE_PRINT_TEXT, true);
        $device->printBarcode();

        $this->_waitWhilePrinting();
        if ($errResult = $this->_isDeviceFailed(__METHOD__)) {
            $this->log(__METHOD__ . " ----- ERROR " . print_r($errResult, true) . " \n", 1);
            return $errResult;
        }

        $this->log(__METHOD__ . " ----- end. OK\n", 3);

        return ['success' => true];
    }

    /**
     * @param FPAtol10Native $device
     * @param $name
     */
    private function setOperatorName($device = null, $name = null)
    {
        if (isset($name) && !empty($name)) {
            $device = isset($device) ? $device : $this->get_NativeDevice();
            $device->setParam(1021, iconv('utf-8', 'windows-1251', $name)); // имя оператора
//            $device->setParam(1203, $this->getInn($device)); // инн
            $device->operatorLogin();
        }
    }

    private function createSession24BlockResponse()
    {
        return [
            'success' => false,
            'error' => 'Текущая смена длится более 24 часов. Для продолжения работы выполните процедуру закрытия смены',
            'errcode' => EC_SESSION_24HOURS_BLOCK,
            'ecrmode' => null,//$device->Mode,
            'ecradvmode' => null,//$device->AdvancedMode
        ];
    }

    private function printOrganizationBatch($params)
    {
        if (isset($params['organizationName']) && strlen($params['organizationName']) > 0) {
            $this->_do_print(['string' => $params['organizationName']]);
        }

        if (isset($params['organizationAddress'])) {
            foreach ($params['organizationAddress'] as $value) {
                $this->_do_print(['string' => $value]);
            }
        }

        if (isset($params['operatorName']) && strlen($params['operatorName']) > 0) {
            $this->_do_print(['string' => $params['operatorName']]);
        }
    }

    /**
     * @param FPAtol10Native $device
     * @param string $receiptType
     * @return array
     */
    private function getVatBunch($device, $receiptType)
    {
        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_SHIFT_TAX_SUM);
        $device->setParam($device->LIBFPTR_PARAM_RECEIPT_TYPE, $device->{$receiptType});
        $device->setParam($device->LIBFPTR_PARAM_TAX_TYPE, $device->LIBFPTR_TAX_VAT10);
        $device->queryData();
        $result['vat10'] = $device->getParamDouble($device->LIBFPTR_PARAM_SUM);

        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_SHIFT_TAX_SUM);
        $device->setParam($device->LIBFPTR_PARAM_RECEIPT_TYPE, $device->{$receiptType});
        $device->setParam($device->LIBFPTR_PARAM_TAX_TYPE, $device->LIBFPTR_TAX_VAT20);
        $device->queryData();
        $result['vat18'] = $device->getParamDouble($device->LIBFPTR_PARAM_SUM);

        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_SHIFT_TAX_SUM);
        $device->setParam($device->LIBFPTR_PARAM_RECEIPT_TYPE, $device->{$receiptType});
        $device->setParam($device->LIBFPTR_PARAM_TAX_TYPE, $device->LIBFPTR_TAX_VAT20);
        $device->queryData();
        $result['vat20'] = $device->getParamDouble($device->LIBFPTR_PARAM_SUM);

        return $result;
    }

    private function getVatBunchDelta($batchBefore, $batchAfter)
    {
        $result['vat10'] = $batchAfter['vat10'] - $batchBefore['vat10'];
        $result['vat18'] = $batchAfter['vat18'] - $batchBefore['vat18'];
        $result['vat20'] = $batchAfter['vat20'] - $batchBefore['vat20'];
        return $result;
    }

    /**
     * @param FPAtol10Native $device
     * @return int
     */
    private function getMoneyProfit($device)
    {
        if (!$this->doJsonQuery($device, ['type' => 'getShiftTotals'], $shiftTotals)) {
            return 0;
        }
        $shiftNumber = $shiftTotals['shiftTotals']['shiftNumber'];
        $sellSum = $shiftTotals['shiftTotals']['receipts']['sell']['sum'];
        $buySum = $shiftTotals['shiftTotals']['receipts']['buy']['sum'];
        $sellReturnSum = $shiftTotals['shiftTotals']['receipts']['sellReturn']['sum'];
        $buyReturnSum = $shiftTotals['shiftTotals']['receipts']['buyReturn']['sum'];
        $revenue = $sellSum - $sellReturnSum - $buySum + $buyReturnSum;

        $this->log(__METHOD__ . " ----- Выручка: $revenue Смена: $shiftNumber\n", 3);

        return $revenue;
    }

    /**
     * @param FPAtol10Native $device
     * @return mixed
     */
    private function getFiscalCheckNumber($device = null)
    {
        $device->setParam($device->LIBFPTR_PARAM_FN_DATA_TYPE, $device->LIBFPTR_FNDT_SHIFT);
        $device->fnQueryData();

        return $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_NUMBER);
    }

    /**
     * @param FPAtol10Native $device
     * @return mixed
     */
    private function getFFDVersion($device = null)
    {
        if (!isset($this->FFDVersion)) {
            $device->setParam($device->LIBFPTR_PARAM_FN_DATA_TYPE, $device->LIBFPTR_FNDT_REG_INFO);
            $device->fnQueryData();
            $this->FFDVersion = (int)$device->getParamInt(1209);
        }
        return $this->FFDVersion;
    }

    /**
     * @param FPAtol10Native $device
     * @return mixed
     */
    private function getLastDocumentNumber($device = null)
    {
        return $this->getStatus($device)['documentNumber'];
    }

    /**
     * @param FPAtol10Native $device
     * @return mixed
     */
    private function getShiftNumber($device = null)
    {
        return $this->getShiftState($device)['number'];
    }

    private function getInn($device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $device->setParam($device->LIBFPTR_PARAM_FN_DATA_TYPE, $device->LIBFPTR_FNDT_TAG_VALUE);
        $device->setParam($device->LIBFPTR_PARAM_TAG_NUMBER, 1018);
        $device->fnQueryData();

        $result = $device->getParamString($device->LIBFPTR_PARAM_TAG_VALUE);

        return trim($result);
    }

    /**
     * @param FPAtol10Native|null $device
     * @return bool
     */
    private function isShiftClosed(FPAtol10Native $device = null)
    {
        return $this->getShiftState($device)['state'] === $device->LIBFPTR_SS_CLOSED;
    }

    /**
     * @param FPAtol10Native|null $device
     * @return bool
     */
    private function isShiftExpired(FPAtol10Native $device = null)
    {
        return $this->getShiftState($device)['state'] === $device->LIBFPTR_SS_EXPIRED;
    }

    /**
     * @param FPAtol10Native|null $device
     * @return array
     */
    private function getShiftState(FPAtol10Native $device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();

        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_SHIFT_STATE);
        $device->queryData();
        return [
            'state' => $device->getParamInt($device->LIBFPTR_PARAM_SHIFT_STATE),
            'number' => $device->getParamInt($device->LIBFPTR_PARAM_SHIFT_NUMBER),
            'dateTime' => (string)$device->getParamDateTime($device->LIBFPTR_PARAM_DATE_TIME),
        ];
    }

    /**
     * @param FPAtol10Native|null $device
     * @return array
     */
    private function getStatus(FPAtol10Native $device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_STATUS);
        $device->queryData();

        return [
            'operatorID' => $device->getParamInt($device->LIBFPTR_PARAM_OPERATOR_ID),
            'logicalNumber' => $device->getParamInt($device->LIBFPTR_PARAM_LOGICAL_NUMBER),
            'shiftState' => $device->getParamInt($device->LIBFPTR_PARAM_SHIFT_STATE),
            'model' => $device->getParamInt($device->LIBFPTR_PARAM_MODEL),
            'mode' => $device->getParamInt($device->LIBFPTR_PARAM_MODE),
            'submode' => $device->getParamInt($device->LIBFPTR_PARAM_SUBMODE),
            'receiptNumber' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_NUMBER),
            'documentNumber' => $device->getParamInt($device->LIBFPTR_PARAM_DOCUMENT_NUMBER),
            'shiftNumber' => $device->getParamInt($device->LIBFPTR_PARAM_SHIFT_NUMBER),
            'receiptType' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_TYPE),
            'lineLength' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_LINE_LENGTH),
            'lineLengthPix' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_LINE_LENGTH_PIX),

            'receiptSum' => $device->getParamDouble($device->LIBFPTR_PARAM_RECEIPT_SUM),

            'isFiscalDevice' => $device->getParamInt($device->LIBFPTR_PARAM_FISCAL),
            'isFiscalFN' => $device->getParamInt($device->LIBFPTR_PARAM_FN_FISCAL),
            'isFNPresent' => $device->getParamInt($device->LIBFPTR_PARAM_FN_PRESENT),
            'isInvalidFN' => $device->getParamInt($device->LIBFPTR_PARAM_INVALID_FN),
            'isCashDrawerOpened' => $device->getParamInt($device->LIBFPTR_PARAM_CASHDRAWER_OPENED),
            'isPaperPresent' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_PAPER_PRESENT),
            'isPaperNearEnd' => $device->getParamInt($device->LIBFPTR_PARAM_PAPER_NEAR_END),
            'isCoverOpened' => $device->getParamInt($device->LIBFPTR_PARAM_COVER_OPENED),
            'isPrinterConnectionLost' => $device->getParamInt($device->LIBFPTR_PARAM_PRINTER_CONNECTION_LOST),
            'isPrinterError' => $device->getParamInt($device->LIBFPTR_PARAM_PRINTER_ERROR),
            'isCutError' => $device->getParamInt($device->LIBFPTR_PARAM_CUT_ERROR),
            'isPrinterOverheat' => $device->getParamInt($device->LIBFPTR_PARAM_PRINTER_OVERHEAT),
            'isDeviceBlocked' => $device->getParamInt($device->LIBFPTR_PARAM_BLOCKED),

            'dateTime' => (string)$device->getParamDateTime($device->LIBFPTR_PARAM_DATE_TIME),

            'serialNumber' => $device->getParamString($device->LIBFPTR_PARAM_SERIAL_NUMBER),
            'modelName' => iconv(
                'windows-1251',
                'utf-8',
                $device->getParamString($device->LIBFPTR_PARAM_MODEL_NAME)
            ),
            'firmwareVersion' => $device->getParamString($device->LIBFPTR_PARAM_UNIT_VERSION),
        ];
    }

    /**
     * @param FPAtol10Native|null $device
     * @return array
     */
    private function getShortStatus(FPAtol10Native $device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_SHORT_STATUS);
        $device->queryData();

        return [
            'isCashDrawerOpened'      => $device->getParamInt($device->LIBFPTR_PARAM_CASHDRAWER_OPENED),
            'isPaperPresent'          => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_PAPER_PRESENT),
            'isPaperNearEnd'          => $device->getParamInt($device->LIBFPTR_PARAM_PAPER_NEAR_END),
            'isCoverOpened'           => $device->getParamInt($device->LIBFPTR_PARAM_COVER_OPENED),
        ];
    }

    /**
     * @param FPAtol10Native|null $device
     * @return array
     */
    private function getReceiptState(FPAtol10Native $device = null)
    {
        $device = isset($device) ? $device : $this->get_NativeDevice();
        $device->setParam($device->LIBFPTR_PARAM_DATA_TYPE, $device->LIBFPTR_DT_RECEIPT_STATE);
        $device->queryData();

        return [
            'receiptType' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_TYPE),
            'receiptNumber' => $device->getParamInt($device->LIBFPTR_PARAM_RECEIPT_NUMBER),
            'documentNumber' => $device->getParamInt($device->LIBFPTR_PARAM_DOCUMENT_NUMBER),
            'sum' => $device->getParamDouble($device->LIBFPTR_PARAM_RECEIPT_SUM),
            'remainder' => $device->getParamDouble($device->LIBFPTR_PARAM_REMAINDER),
            'change' => $device->getParamDouble($device->LIBFPTR_PARAM_CHANGE),
        ];
    }

    private function finalDiscountConvertCheckPrint($lines)
    {
        $newLines = [];
        foreach ($lines as $line) {
            $action = $line['action'];
            if ($action === "discount") {
                $newLine['action'] = "print";
                $description = $line['params']['description'];
                $finalDiscountAmount = number_format($line['params']['amount'], 2);
                if (!empty($description) && $description !== ' ') {
                    $newLine['params']['string'] = "$description" . ": " . $finalDiscountAmount;
                } else {
                    $newLine['params']['string'] = "СКИДКА: " . $finalDiscountAmount;
                }
                $newLine['params']['alignment'] = 2; // выравниваем по правому краю, работает только в протоколе АТОЛ 2.x
                $newLines[] = $newLine;
            } else {
                $newLines[] = $line;
            }
        }
        return $newLines;
    }
}
